/*
 * Copyright (c) 2011 ScientiaMobile, Inc., VA, USA
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * Refer to the COPYING file distributed with this package.
 */
package net.sourceforge.wurfl.core.handlers.matchers.strategy;

import java.util.Iterator;
import java.util.List;
import java.util.SortedSet;

import org.apache.commons.lang.text.StrBuilder;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * Token matcher
 * 
 * <p>
 * This matcher assign to user-agent token different weight. Work with well know
 * user-agent patterns. It can be useful with Os or firmware version number. The
 * {@link TokensProvider} interface provide tokens and respective weights.
 * </p>
 * 
 * @author WURFL-PRO SRL, Rome, Italy
 * @version $Id: TokensMatcher.java 1045 2009-03-09 15:52:53Z filippo.deluca $
 */
public class TokensMatcher implements StringMatcher {

	/** Log */
	private static final Log log = LogFactory.getLog(TokensMatcher.class);

	/** The TokensProvider used to get Tokens from user-agent. */
	private TokensProvider tokensProvider;

	/**
	 * Build new TokenMatcher by TokensProvider
	 * 
	 * @param tokensProvider
	 *            TokensProvider used to tokenize User-Agent.
	 */
	public TokensMatcher(TokensProvider tokensProvider) {
		this.tokensProvider = tokensProvider;
	}

	public String match(SortedSet userAgentsSet, String needle, int tollerance) {

		if (log.isTraceEnabled()) {
			StrBuilder sb = new StrBuilder("Applying TK(").append(tollerance)
					.append(") on User-Agent: ").append(needle).append(
							" using User-Agent set: [");

			for (Iterator uIt = userAgentsSet.iterator(); uIt.hasNext();) {
				sb.append(uIt.next());

				if (uIt.hasNext()) {
					sb.append(", ");
				}
			}
			sb.append("]");

			log.trace(sb.toString());
		}

		List needleGroups = tokensProvider.createTokens(needle);
		assert needleGroups != null : "needleGroups is null";

		// Tokenize needle
		int needleGroupsCount = needleGroups.size();
		Iterator iterator = userAgentsSet.iterator();
		String bestMatch = null;
		int bestDiff = tollerance + 1;

		while (iterator.hasNext()) {

			String userAgent = (String) iterator.next();

			// Tokenize User-Agent
			List userAgentTokens = tokensProvider.createTokens(userAgent);
			assert userAgentTokens != null : "userAgentGroups is null";
			int userAgentTokensCount = userAgentTokens.size();

			// Logs tokens
			if (log.isTraceEnabled()) {
				StrBuilder logBuilder = new StrBuilder();
				logBuilder.append("userAgentTokens: [");

				for (int i = 0; i < userAgentTokensCount; i++) {
					logBuilder.append(userAgentTokens.get(i));
					if (i < userAgentTokensCount - 1) {
						logBuilder.append(", ");
					}
				}

				logBuilder.append("] needleTokens: [");

				for (int i = 0; i < needleGroupsCount; i++) {
					logBuilder.append(needleGroups.get(i));
					if (i < needleGroupsCount - 1) {
						logBuilder.append(", ");
					}
				}

				logBuilder.append("]");

				log.trace(logBuilder.toString());
			}

			if (userAgentTokensCount == needleGroupsCount) {

				int weightDiff = 0;

				for (int groupIndex = 0; groupIndex < userAgentTokensCount
						&& weightDiff < tollerance; groupIndex++) {
					boolean equals = areEquals((String) needleGroups
							.get(groupIndex), (String) userAgentTokens
							.get(groupIndex));

					if (!equals) {
						weightDiff += tokensProvider.getTokenWeight(groupIndex);
					}
				}

				if (weightDiff < bestDiff) {

					bestDiff = weightDiff;
					bestMatch = userAgent;

					// Create log
					if (log.isTraceEnabled()) {
						StrBuilder sb = new StrBuilder();
						sb.append("Found new best match ");
						sb.append("User-Agent: ").append(userAgent).append(" ");
						sb.append("needle: ").append(needle).append(" ");
						sb.append("diff: ").append(weightDiff);

						log.trace(sb.toString());
					}

				}

			}
		}

		return bestMatch;
	}

	protected boolean areEquals(String s1, String s2) {
		return s1.equals(s2);
	}

}
