/*
 * Copyright (c) 2011 ScientiaMobile, Inc., VA, USA
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Affero General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * Refer to the COPYING file distributed with this package.
 */
package net.sourceforge.wurfl.core.resource;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URISyntaxException;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.zip.GZIPInputStream;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.apache.commons.lang.StringUtils;
import org.apache.commons.lang.SystemUtils;
import org.apache.commons.lang.Validate;
import org.apache.commons.lang.text.StrBuilder;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.xml.sax.Attributes;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

/**
 * XMLResource
 * 
 * <p>
 * XMLResource represent a source of wurfl repository backed by XML file. XML
 * file can be root file or patch. The XML file can be compresses by gzip or
 * zip. In case of zip, the first entry will be processed.
 * </p>
 * 
 * <p>
 * The given path can be URL or filesystem path. The URL support classpath
 * scheme to load file from java classpath.
 * </p>
 * 
 * @author WURFL-PRO SRL, Rome, Italy
 * @version $Id: XMLResource.java 1045 2009-03-09 15:52:53Z filippo.deluca $
 */
public class XMLResource implements WURFLResource {

	/** Logger */
	private static final Log log = LogFactory.getLog(XMLResource.class);

	/** Source File */
	private File source;

	/**
	 * Build Resource by path. The path can be URL or filesystem path.
	 * 
	 * @param path
	 *            The path of source file.
	 */
	public XMLResource(String path) {

		Validate.notEmpty(path, "The path is empty");

		File file = createFile(path);

		init(file);
	}

	/**
	 * Build resource by File.
	 * 
	 * @param file
	 *            The source File
	 */
	public XMLResource(File file) {

		init(file);
	}
	
	// Access methods *****************************************************

	/**
	 * {@inheritDoc}
	 */
	public ResourceData getData() {
		// TODO Auto-generated method stub

		WURFLSAXHandler handler = new WURFLSAXHandler();

		try {

			InputStream input = openInputStream();

			SAXParser parser = SAXParserFactory.newInstance().newSAXParser();
			parser.parse(input, handler);

			input.close();

		} catch (Exception e) {
			throw new WURFLResourceException(this, e);
		}

		String name = getInfo();
		String version = handler.getVer();
		boolean patch = handler.isPatch();
		ModelDevices devices = handler.getDevices();

		ResourceData data = new ResourceData(name, version, patch, devices);

		return data;
	}

	/**
	 * Open InputStream from source file.
	 * 
	 * @return InputStream obtained from the source File.
	 * 
	 * @throws IOException
	 *             If an exception occured while opening the source file.
	 */
	protected InputStream openInputStream() throws IOException {

		InputStream input = null;

		if (source.getName().toLowerCase().endsWith(".zip")) {
			ZipFile zipFile = new ZipFile(source);
			ZipEntry zipEntry = (ZipEntry) zipFile.entries().nextElement();
			input = zipFile.getInputStream(zipEntry);
		} else if (source.getName().toLowerCase().endsWith(".gz")) {
			input = new GZIPInputStream(new FileInputStream(source));
		} else {
			input = new FileInputStream(source);
		}

		return input;
	}

	/**
	 * {@inheritDoc}
	 */
	public String getInfo() {

		String info = null;

		try {
			info = source.getCanonicalPath();
		} catch (IOException e) {
			info = source.getAbsolutePath();
			log.warn("Error getting path for file: " + info);
		}

		return info;
	}

	/**
	 * {@inheritDoc}
	 */
	public void release() {
		source = null;
	}

	// Helper methods *****************************************************

	/**
	 * Convert spaces to &quot;%20&quot;
	 * 
	 * @param location
	 *            Location to convert.
	 * @return converted location.
	 */
	public static URI toURI(String location) throws URISyntaxException {
		return new URI(StringUtils.replace(location, " ", "%20"));
	}

	/**
	 * Create source File by path
	 * 
	 * @param path
	 *            The path of source File.
	 * @return The File obtained by path.
	 */
	protected File createFile(String path) {

		File file = null;

		try {
			URI uri = null;
			
			// FIXME where the path must be normalized?
			if(SystemUtils.IS_OS_WINDOWS && StringUtils.contains(path, "\\")){
				log.debug("Encoding windows URL");
				path = path.replaceAll("\\\\", "/");
				path = "file:///" + path;
			}

			if (path.startsWith("classpath:")) {
				path = path.substring("classpath:".length());
				uri = new URI(WURFLResource.class.getResource(path).toString());
			} else if (path.indexOf(":") != -1) {
				uri = new URI(path);
			} else {
				File tmp = new File(path);
				uri = tmp.toURI();
			}

			log.debug("Creating file from path: " + uri.toString());
			file = new File(uri);

		} catch (Exception e) {
			throw new WURFLResourceException(this, e);
		}

		return file;
	}

	/**
	 * Init method
	 * 
	 * @param source
	 *            The XML source file.
	 */
	protected void init(File source) {

		Validate.notNull(source);
		Validate.isTrue(source.exists(), "File: " + source.getAbsolutePath()
				+ " does not exist");
		Validate.isTrue(source.canRead());

		this.source = source;
	}

	// Helper classes *****************************************************

	static class WURFLSAXHandler extends DefaultHandler {

		// XML costants ***************************************************

		private static final String ELEM_WURFL = "wurfl";
		private static final String ELEM_WURFL_PATCH = "wurfl_patch";
		private static final String ELEM_DEVICE = "device";
		private static final String ELEM_DEVICES = "devices";
		private static final String ELEM_GROUP = "group";
		private static final String ELEM_CAPABILITY = "capability";
		private static final String ELEM_VERSION = "version";
		private static final String ELEM_VER = "ver";

		private static final String ATTR_DEVICE_ID = "id";
		private static final String ATTR_DEVICE_FALLBACK = "fall_back";
		private static final String ATTR_DEVICE_USERAGENT = "user_agent";
		private static final String ATTR_DEVICE_ACTUALDEVICEROOT = "actual_device_root";

		private static final String ATTR_GROUP_ID = "id";

		private static final String ATTR_CAPABILITY_NAME = "name";
		private static final String ATTR_CAPABILITY_VALUE = "value";

		// Memebers *******************************************************

		private String userAgent;

		private String deviceID;

		private String fallBack;

		private boolean actualDeviceRoot;

		private String groupID;

		private String capabilityName;

		private String capabilityValue;

		private Set userAgents;

		private Set devicesId;

		private Map capabilities;

		private Map capabilitiesByGroup;

		private ModelDevices devices;

		private String ver;

		private boolean insideVer = false;

		private boolean patch = false;

		private boolean root = false;

		// Access methods *************************************************

		public String getVer() {
			return ver;
		}

		public ModelDevices getDevices() {
			return devices;
		}

		public boolean isPatch() {
			return patch;
		}

		// Parser methods *************************************************

		public void startDocument() throws SAXException {
			super.startDocument();
			userAgents = new HashSet();
			devicesId = new HashSet();
			devices = new ModelDevices();
		}

		public void endDocument() throws SAXException {
			// Check if all devices have a valid fallback (only for root WURFL)
			// for (Iterator iterator = devicesMap.values().iterator();
			// iterator.hasNext();) {
			// ModelDevice device = (ModelDevice) iterator.next();
			// // Only for root WURFL
			// if (!"generic".equals(device.getID())) {
			// if (!devicesMap.containsKey(device.getFallBack())) {
			// StringBuffer msg = new StringBuffer();
			// msg.append("Device with device id ").append(device.getID()).append(
			// " has invalid fall back value ").append(device.getFallBack());
			// throw new WURFLParsingException(msg.toString());
			// }
			// }
			// }
		}

		public void startElement(String uri, String localName, String name,
				Attributes attributes) throws SAXException {
			if (ELEM_DEVICE.equals(name)) {
				deviceStartElement(attributes);
			} else if (ELEM_GROUP.equals(name)) {
				startGroupElement(attributes);
			} else if (ELEM_CAPABILITY.equals(name)) {
				startCapabilityElement(attributes);
			} else if (ELEM_VER.equals(name)) {
				startVerElement();
			} else if (ELEM_WURFL_PATCH.equals(name)) {
				startWurflPatchElement();
			} else if (ELEM_WURFL.equals(name)) {
				startWurflElement();
			}
		}

		public void endElement(String uri, String localName, String name)
				throws SAXException {
			if (ELEM_GROUP.equals(name)) {
				endGroupElement();
			} else if (ELEM_DEVICE.equals(name)) {
				endDeviceElement();
			} else if (ELEM_VER.equals(name)) {
				endVerElement();
			}
		}

		public void characters(char[] ch, int start, int length)
				throws SAXException {
			if (insideVer) {
				charsVer(ch, start, length);
			}
		}

		private void startWurflElement() {

			if (patch || root) {
				throw new WURFLParsingException(
						"Root element already defined: wurfl");
			}

			root = true;

		}

		private void startWurflPatchElement() {
			if (patch || root) {
				throw new WURFLParsingException(
						"Root element already defined: wurfl_patch");
			}

			patch = true;
		}

		private void startVerElement() {
			insideVer = true;
		}

		public void charsVer(char[] ch, int start, int length) {
			StrBuilder verBuilder = new StrBuilder();
			verBuilder.append(ch, start, length);

			ver = verBuilder.toString();
		}

		private void endVerElement() {
			insideVer = false;
		}

		private void deviceStartElement(Attributes attributes) {

			userAgent = attributes.getValue(ATTR_DEVICE_USERAGENT);
			deviceID = attributes.getValue(ATTR_DEVICE_ID);
			fallBack = attributes.getValue(ATTR_DEVICE_FALLBACK);
			actualDeviceRoot = Boolean.valueOf(
					attributes.getValue(ATTR_DEVICE_ACTUALDEVICEROOT))
					.booleanValue();

			// check if the user agent and device id are valid values
			if (StringUtils.isEmpty(deviceID)) {
				throw new WURFLParsingException("device id is not a valid");
			}
			if (!"generic".equals(deviceID) && StringUtils.isEmpty(userAgent)) {
				StringBuffer msg = new StringBuffer();
				msg.append("Device with id ").append(deviceID).append(
						" has an invalid user agent");
				throw new WURFLParsingException(msg.toString());
			}

			// check if the device id and the user_agent are unique
			if (devicesId.contains(deviceID)) {
				throw new WURFLParsingException("device id " + deviceID
						+ " already defined!!!");
			}

			if (userAgents.contains(userAgent)) {
				throw new WURFLParsingException("user agent [" + userAgent
						+ "] already defined");
			}

			userAgents.add(userAgent);
			devicesId.add(deviceID);

			capabilities = new HashMap();
			capabilitiesByGroup = new HashMap();
		}

		private void endDeviceElement() {

			ModelDevice device = new ModelDevice.Builder(deviceID, userAgent,
					fallBack).setActualDeviceRoot(actualDeviceRoot)
					.setCapabilities(capabilities).setCapabilitiesByGroup(
							capabilitiesByGroup).build();

			devices.add(device);
		}

		private void startCapabilityElement(Attributes attributes) {

			capabilityName = attributes.getValue(ATTR_CAPABILITY_NAME);
			capabilityValue = attributes.getValue(ATTR_CAPABILITY_VALUE);

			if (StringUtils.isEmpty(capabilityName) || null == capabilityValue) {
				throw new WURFLParsingException("device with id " + deviceID
						+ " has capability with name or value not valid");
			}

			if (capabilities.containsKey(capabilityName)) {
				throw new WURFLParsingException("The devices with id "
						+ deviceID + " define more " + capabilityName);
			}

			// intern by Kenny MacLeod
			String internCapabilityName = capabilityName.intern();
			String internGroupId = groupID.intern();

			capabilities.put(internCapabilityName, capabilityValue);
			capabilitiesByGroup.put(internCapabilityName, internGroupId);
		}

		private void startGroupElement(Attributes attributes) {

			groupID = attributes.getValue(ATTR_GROUP_ID);

		}

		private void endGroupElement() {
			/* Empty */
		}

	}

}
